import logging

from django.apps import apps

from django_pgviews.signals import view_synced, all_views_synced
from django_pgviews.view import create_view, View, MaterializedView, create_materialized_view

logger = logging.getLogger("django_pgviews.sync_pgviews")
exists_logger = logging.getLogger("django_pgviews.sync_pgviews.exists")


class RunBacklog(object):
    def __init__(self) -> None:
        super().__init__()
        self.finished = []

    def run(self, **kwargs):
        self.finished = []
        backlog = []
        for view_cls in apps.get_models():
            if not (isinstance(view_cls, type) and issubclass(view_cls, View) and hasattr(view_cls, "sql")):
                continue
            backlog.append(view_cls)
        loop = 0
        while len(backlog) > 0 and loop < 10:
            loop += 1
            backlog = self.run_backlog(backlog, **kwargs)

        if loop >= 10:
            logger.warning("pgviews dependencies hit limit. Check if your model dependencies are correct")
            return False

        return True

    def run_backlog(self, backlog, **kwargs):
        raise NotImplementedError


class ViewSyncer(RunBacklog):
    def run(self, force, update, using, materialized_views_check_sql_changed=False, **options):
        if super().run(
            force=force,
            update=update,
            using=using,
            materialized_views_check_sql_changed=materialized_views_check_sql_changed,
        ):
            all_views_synced.send(sender=None, using=using)

    def run_backlog(self, backlog, *, force, update, using, materialized_views_check_sql_changed, **kwargs):
        """Installs the list of models given from the previous backlog

        If the correct dependent views have not been installed, the view
        will be added to the backlog.

        Eventually we get to a point where all dependencies are sorted.
        """
        new_backlog = []
        for view_cls in backlog:
            skip = False
            name = "{}.{}".format(view_cls._meta.app_label, view_cls.__name__)
            for dep in view_cls._dependencies:
                if dep not in self.finished:
                    skip = True
                    break

            if skip is True:
                new_backlog.append(view_cls)
                logger.info("Putting pgview at back of queue: %s", name)
                continue  # Skip

            try:
                connection = view_cls.get_view_connection(using=using, restricted_mode=True)
                if not connection:
                    logger.info("Skipping pgview %s (migrations not allowed on %s)", name, using)
                    continue  # Skip
                if isinstance(view_cls(), MaterializedView):
                    status = create_materialized_view(
                        connection, view_cls, check_sql_changed=materialized_views_check_sql_changed
                    )
                else:
                    status = create_view(
                        connection,
                        view_cls._meta.db_table,
                        view_cls.get_sql(),
                        update=update,
                        force=force,
                    )

                view_synced.send(
                    sender=view_cls,
                    update=update,
                    force=force,
                    status=status,
                    has_changed=status not in ("EXISTS", "FORCE_REQUIRED"),
                    using=using,
                )
                self.finished.append(name)
            except Exception as exc:
                exc.view_cls = view_cls
                exc.python_name = name
                raise
            else:
                use_logger = logger

                if status == "CREATED":
                    msg = "created"
                elif status == "UPDATED":
                    msg = "updated"
                elif status == "EXISTS":
                    msg = "already exists, skipping"
                    use_logger = exists_logger
                elif status == "FORCED":
                    msg = "forced overwrite of existing schema"
                elif status == "FORCE_REQUIRED":
                    msg = "exists with incompatible schema, --force required to update"
                else:
                    msg = status

                use_logger.info("pgview %s %s", name, msg)
        return new_backlog


class ViewRefresher(RunBacklog):
    def run(self, concurrently, using, **kwargs):
        return super().run(concurrently=concurrently, using=using, **kwargs)

    def run_backlog(self, backlog, *, concurrently, using, **kwargs):
        new_backlog = []
        for view_cls in backlog:
            skip = False
            name = "{}.{}".format(view_cls._meta.app_label, view_cls.__name__)
            for dep in view_cls._dependencies:
                if dep not in self.finished:
                    skip = True
                    break

            if skip is True:
                new_backlog.append(view_cls)
                logger.info("Putting pgview at back of queue: %s", name)
                continue  # Skip

            # Don't refresh views not associated with this database
            connection = view_cls.get_view_connection(using=using, restricted_mode=True)
            if not connection:
                continue

            if issubclass(view_cls, MaterializedView):
                view_cls.refresh(concurrently=concurrently)
                logger.info("pgview %s refreshed", name)

            self.finished.append(name)

        return new_backlog



import logging

from django.apps import apps
from django.core.management.base import BaseCommand
from django.db import DEFAULT_DB_ALIAS

from django_pgviews.view import clear_view, View, MaterializedView

logger = logging.getLogger("django_pgviews.sync_pgviews")


class Command(BaseCommand):
    help = """Clear Postgres views. Use this before running a migration"""

    def add_arguments(self, parser):
        parser.add_argument(
            "--database",
            default=DEFAULT_DB_ALIAS,
            help='Nominates a database to synchronize. Defaults to the "default" database.',
        )

    def handle(self, database, **options):
        for view_cls in apps.get_models():
            if not (isinstance(view_cls, type) and issubclass(view_cls, View) and hasattr(view_cls, "sql")):
                continue
            python_name = "{}.{}".format(view_cls._meta.app_label, view_cls.__name__)
            connection = view_cls.get_view_connection(using=database, restricted_mode=True)
            if not connection:
                continue
            status = clear_view(
                connection, view_cls._meta.db_table, materialized=isinstance(view_cls(), MaterializedView)
            )
            if status == "DROPPED":
                msg = "dropped"
            else:
                msg = "not dropped"
            logger.info("%s (%s): %s", python_name, view_cls._meta.db_table, msg)



from django.db import connections
from django.db.models.sql import query

from django_pgviews.db.sql import compiler


class NonQuotingQuery(query.Query):
    """
    Query class that uses the NonQuotingCompiler.
    """

    def get_compiler(self, using=None, connection=None):
        """
        Get the NonQuotingCompiler object.
        """
        if using is None and connection is None:
            raise ValueError("Need either using or connection")
        if using:
            connection = connections[using]

        for alias, annotation in self.annotation_select.items():
            connection.ops.check_expression_support(annotation)

        return compiler.NonQuotingCompiler(self, connection, using)



from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import DEFAULT_DB_ALIAS

from django_pgviews.models import ViewSyncer


class Command(BaseCommand):
    help = """Create/update Postgres views for all installed apps."""

    def add_arguments(self, parser):
        parser.add_argument(
            "--no-update",
            action="store_false",
            dest="update",
            default=True,
            help="""Don't update existing views, only create new ones.""",
        )
        parser.add_argument(
            "--force",
            action="store_true",
            dest="force",
            default=False,
            help="Force replacement of pre-existing views where breaking changes have been made to the schema.",
        )
        parser.add_argument(
            "-E",
            "--enable-materialized-views-check-sql-changed",
            action="store_true",
            dest="materialized_views_check_sql_changed",
            default=None,
            help=(
                "Before recreating materialized view, check the SQL has changed compared to the currently active "
                "materialized view in the database, if there is one, and only re-create the materialized view "
                "if the SQL is different. By default uses django setting MATERIALIZED_VIEWS_CHECK_SQL_CHANGED."
            ),
        )
        parser.add_argument(
            "-D",
            "--disable-materialized-views-check-sql-changed",
            action="store_false",
            dest="materialized_views_check_sql_changed",
            default=None,
            help=(
                "Before recreating materialized view, check the SQL has changed compared to the currently active "
                "materialized view in the database, if there is one, and only re-create the materialized view "
                "if the SQL is different. By default uses django setting MATERIALIZED_VIEWS_CHECK_SQL_CHANGED."
            ),
        )
        parser.add_argument(
            "--database",
            default=DEFAULT_DB_ALIAS,
            help='Nominates a database to synchronize. Defaults to the "default" database.',
        )

    def handle(self, force, update, materialized_views_check_sql_changed, database, **options):
        vs = ViewSyncer()

        if materialized_views_check_sql_changed is None:
            materialized_views_check_sql_changed = getattr(settings, "MATERIALIZED_VIEWS_CHECK_SQL_CHANGED", False)

        vs.run(force, update, using=database, materialized_views_check_sql_changed=materialized_views_check_sql_changed)



from django.core.management.base import BaseCommand
from django.db import DEFAULT_DB_ALIAS

from django_pgviews.models import ViewRefresher


class Command(BaseCommand):
    help = """Refresh materialized Postgres views for all installed apps."""

    def add_arguments(self, parser):
        parser.add_argument(
            "-C",
            "--concurrently",
            action="store_true",
            dest="concurrently",
            help="Refresh concurrently if the materialized view supports it",
        )
        parser.add_argument(
            "--database",
            default=DEFAULT_DB_ALIAS,
            help='Nominates a database to synchronize. Defaults to the "default" database.',
        )

    def handle(self, concurrently, database, **options):
        ViewRefresher().run(concurrently, using=database)



"""Helpers to access Postgres views from the Django ORM.
"""
import collections
import copy
import logging
import re

import django
from django.apps import apps
from django.core import exceptions
from django.db import connections, router, transaction
from django.db import models
from django.db.backends.postgresql.schema import DatabaseSchemaEditor
from django.db.backends.utils import truncate_name
from django.db.models.query import QuerySet

from django_pgviews.compat import ProgrammingError
from django_pgviews.db import get_fields_by_name


FIELD_SPEC_REGEX = r"^([A-Za-z_][A-Za-z0-9_]*)\." r"([A-Za-z_][A-Za-z0-9_]*)\." r"(\*|(?:[A-Za-z_][A-Za-z0-9_]*))$"
FIELD_SPEC_RE = re.compile(FIELD_SPEC_REGEX)

ViewSQL = collections.namedtuple("ViewSQL", "query,params")

logger = logging.getLogger("django_pgviews.view")


def hasfield(model_cls, field_name):
    """
    Like `hasattr()`, but for model fields.

    >>> from django.contrib.auth.models import User
    >>> hasfield(User, 'password')
    True
    >>> hasfield(User, 'foobarbaz')
    False
    """
    try:
        model_cls._meta.get_field(field_name)
        return True
    except exceptions.FieldDoesNotExist:
        return False


# Projections of models fields onto views which have been deferred due to
# model import and loading dependencies.
# Format: (app_label, model_name): {view_cls: [field_name, ...]}
_DEFERRED_PROJECTIONS = collections.defaultdict(lambda: collections.defaultdict(list))


def realize_deferred_projections(sender, *args, **kwargs):
    """Project any fields which were deferred pending model preparation."""
    app_label = sender._meta.app_label
    model_name = sender.__name__.lower()
    pending = _DEFERRED_PROJECTIONS.pop((app_label, model_name), {})
    for view_cls, field_names in pending.items():
        field_instances = get_fields_by_name(sender, *field_names)
        for name, field in field_instances.items():
            # Only assign the field if the view does not already have an
            # attribute or explicitly-defined field with that name.
            if hasattr(view_cls, name) or hasfield(view_cls, name):
                continue
            copy.copy(field).contribute_to_class(view_cls, name)


models.signals.class_prepared.connect(realize_deferred_projections)


def _schema_and_name(connection, view_name):
    if "." in view_name:
        return view_name.split(".", 1)
    else:
        try:
            schema_name = connection.schema_name
        except AttributeError:
            schema_name = "public"

        return schema_name, view_name


def _create_mat_view(cursor, view_name, query, params, with_data):
    """
    Creates a materialized view using a specific cursor, name and definition.
    """
    cursor.execute(
        "CREATE MATERIALIZED VIEW {0} AS {1} {2};".format(
            view_name, query, "WITH DATA" if with_data else "WITH NO DATA"
        ),
        params,
    )


def _drop_mat_view(cursor, view_name):
    """
    Drops a materialized view using a specific cursor.
    """
    cursor.execute("DROP MATERIALIZED VIEW IF EXISTS {0} CASCADE;".format(view_name))


def _concurrent_index_name(view_name, concurrent_index):
    # replace . with _ in view_name in case the table is in a schema
    return view_name.replace(".", "_") + "_" + "_".join([s.strip() for s in concurrent_index.split(",")]) + "_index"


def _create_concurrent_index(cursor, view_name, concurrent_index):
    cursor.execute(
        "CREATE UNIQUE INDEX {index_name} ON {view_name} ({concurrent_index})".format(
            view_name=view_name,
            index_name=_concurrent_index_name(view_name, concurrent_index),
            concurrent_index=concurrent_index,
        )
    )


class CustomSchemaEditor(DatabaseSchemaEditor):
    def _create_index_sql(self, *args, **kwargs):
        """
        Override to handle indexes in custom schemas, when the schema is explicitly set.
        """
        statement = super()._create_index_sql(*args, **kwargs)

        model = args[0]

        if "." in model._meta.db_table:  # by default the table it's quoted, but we need it non-quoted
            statement.parts["table"] = model._meta.db_table

        return statement


def _ensure_indexes(connection, cursor, view_cls, schema_name_log):
    """
    This function gets called when a materialized view is deemed not needing a re-create. That is however only a part
    of the story, since that checks just the SQL of the view itself. The second part is the indexes.
    This function gets the current indexes on the materialized view and reconciles them with the indexes that
    should be in the view, dropping extra ones and creating new ones.
    """
    view_name = view_cls._meta.db_table
    concurrent_index = view_cls._concurrent_index
    indexes = view_cls._meta.indexes
    vschema, vname = _schema_and_name(connection, view_name)

    cursor.execute("SELECT indexname FROM pg_indexes WHERE tablename = %s AND schemaname = %s", [vname, vschema])

    existing_indexes = set(x[0] for x in cursor.fetchall())
    required_indexes = set(x.name for x in indexes)

    if view_cls._concurrent_index is not None:
        concurrent_index_name = _concurrent_index_name(view_name, concurrent_index)
        required_indexes.add(concurrent_index_name)
    else:
        concurrent_index_name = None

    for index_name in existing_indexes - required_indexes:
        cursor.execute(f"DROP INDEX {vschema}.{index_name}")
        logger.info("pgview dropped index %s on view %s (%s)", index_name, view_name, schema_name_log)

    schema_editor: DatabaseSchemaEditor = CustomSchemaEditor(connection)

    for index_name in required_indexes - existing_indexes:
        if index_name == concurrent_index_name:
            _create_concurrent_index(cursor, view_name, concurrent_index)
            logger.info("pgview created concurrent index on view %s (%s)", view_name, schema_name_log)
        else:
            for index in indexes:
                if index.name == index_name:
                    schema_editor.add_index(view_cls, index)
                    logger.info("pgview created index %s on view %s (%s)", index.name, view_name, schema_name_log)
                    break


@transaction.atomic()
def create_materialized_view(connection, view_cls, check_sql_changed=False):
    """
    Create a materialized view on a connection.

    Returns one of statuses EXISTS, UPDATED, CREATED.

    If with_data = False, then the materialized view will get created without data.

    If check_sql_changed = True, then the process will first check if there is a materialized view in the database
    already with the same SQL, if there is, it will not do anything. Otherwise the materialized view gets dropped
    and recreated.
    """
    view_name = view_cls._meta.db_table
    view_query = view_cls.get_sql()
    concurrent_index = view_cls._concurrent_index

    try:
        schema_name = connection.schema_name
        schema_name_log = f"schema {schema_name}"
    except AttributeError:
        schema_name_log = "default schema"

    vschema, vname = _schema_and_name(connection, view_name)

    cursor_wrapper = connection.cursor()
    cursor = cursor_wrapper.cursor

    try:
        cursor.execute(
            "SELECT COUNT(*) FROM pg_matviews WHERE schemaname = %s and matviewname = %s;",
            [vschema, vname],
        )
        view_exists = cursor.fetchone()[0] > 0

        query = view_query.query.strip()
        if query.endswith(";"):
            query = query[:-1]

        if check_sql_changed and view_exists:
            temp_viewname = truncate_name(view_name + "_temp", length=63)
            _, temp_vname = _schema_and_name(connection, temp_viewname)

            _drop_mat_view(cursor, temp_viewname)
            _create_mat_view(cursor, temp_viewname, query, view_query.params, with_data=False)

            cursor.execute(
                "SELECT definition FROM pg_matviews WHERE schemaname = %s and matviewname IN (%s, %s);",
                [vschema, vname, temp_vname],
            )
            definitions = cursor.fetchall()

            _drop_mat_view(cursor, temp_viewname)

            if definitions[0] == definitions[1]:
                _ensure_indexes(connection, cursor, view_cls, schema_name_log)
                return "EXISTS"

        if view_exists:
            _drop_mat_view(cursor, view_name)
            logger.info("pgview dropped materialized view %s (%s)", view_name, schema_name_log)

        _create_mat_view(cursor, view_name, query, view_query.params, with_data=view_cls.with_data)
        logger.info("pgview created materialized view %s (%s)", view_name, schema_name_log)

        if concurrent_index is not None:
            _create_concurrent_index(cursor, view_name, concurrent_index)
            logger.info("pgview created concurrent index on view %s (%s)", view_name, schema_name_log)

        if view_cls._meta.indexes:
            schema_editor = CustomSchemaEditor(connection)

            for index in view_cls._meta.indexes:
                schema_editor.add_index(view_cls, index)
                logger.info("pgview created index %s on view %s (%s)", index.name, view_name, schema_name_log)

        if view_exists:
            return "UPDATED"

        return "CREATED"
    finally:
        cursor_wrapper.close()


@transaction.atomic()
def create_view(connection, view_name, view_query: ViewSQL, update=True, force=False):
    """
    Create a named view on a connection.

    Returns True if a new view was created (or an existing one updated), or
    False if nothing was done.

    If ``update`` is True (default), attempt to update an existing view. If the
    existing view's schema is incompatible with the new definition, ``force``
    (default: False) controls whether or not to drop the old view and create
    the new one.
    """

    vschema, vname = _schema_and_name(connection, view_name)

    cursor_wrapper = connection.cursor()
    cursor = cursor_wrapper.cursor
    try:
        force_required = False
        # Determine if view already exists.
        cursor.execute(
            "SELECT COUNT(*) FROM information_schema.views WHERE table_schema = %s and table_name = %s;",
            [vschema, vname],
        )
        view_exists = cursor.fetchone()[0] > 0
        if view_exists and not update:
            return "EXISTS"
        elif view_exists:
            # Detect schema conflict by copying the original view, attempting to
            # update this copy, and detecting errors.
            cursor.execute("CREATE TEMPORARY VIEW check_conflict AS SELECT * FROM {0};".format(view_name))
            try:
                with transaction.atomic():
                    cursor.execute(
                        "CREATE OR REPLACE TEMPORARY VIEW check_conflict AS {0};".format(view_query.query),
                        view_query.params,
                    )
            except ProgrammingError:
                force_required = True
            finally:
                cursor.execute("DROP VIEW IF EXISTS check_conflict;")

        if not force_required:
            cursor.execute("CREATE OR REPLACE VIEW {0} AS {1};".format(view_name, view_query.query), view_query.params)
            ret = view_exists and "UPDATED" or "CREATED"
        elif force:
            cursor.execute("DROP VIEW IF EXISTS {0} CASCADE;".format(view_name))
            cursor.execute("CREATE VIEW {0} AS {1};".format(view_name, view_query.query), view_query.params)
            ret = "FORCED"
        else:
            ret = "FORCE_REQUIRED"

        return ret
    finally:
        cursor_wrapper.close()


def clear_view(connection, view_name, materialized=False):
    """
    Remove a named view on connection.
    """
    cursor_wrapper = connection.cursor()
    cursor = cursor_wrapper.cursor
    try:
        if materialized:
            cursor.execute("DROP MATERIALIZED VIEW IF EXISTS {0} CASCADE".format(view_name))
        else:
            cursor.execute("DROP VIEW IF EXISTS {0} CASCADE".format(view_name))
    finally:
        cursor_wrapper.close()
    return "DROPPED".format(view=view_name)


class ViewMeta(models.base.ModelBase):
    def __new__(metacls, name, bases, attrs):
        """
        Deal with all of the meta attributes, removing any Django does not want
        """
        # Get attributes before Django
        dependencies = attrs.pop("dependencies", [])
        projection = attrs.pop("projection", [])
        concurrent_index = attrs.pop("concurrent_index", None)

        # Get projection
        deferred_projections = []
        for field_name in projection:
            if isinstance(field_name, models.Field):
                attrs[field_name.name] = copy.copy(field_name)
            elif isinstance(field_name, str):
                match = FIELD_SPEC_RE.match(field_name)
                if not match:
                    raise TypeError("Unrecognized field specifier: %r" % field_name)
                deferred_projections.append(match.groups())
            else:
                raise TypeError("Unrecognized field specifier: %r" % field_name)

        view_cls = models.base.ModelBase.__new__(metacls, name, bases, attrs)

        # Get dependencies
        setattr(view_cls, "_dependencies", dependencies)
        # Materialized views can have an index allowing concurrent refresh
        setattr(view_cls, "_concurrent_index", concurrent_index)
        for app_label, model_name, field_name in deferred_projections:
            model_spec = (app_label, model_name.lower())

            _DEFERRED_PROJECTIONS[model_spec][view_cls].append(field_name)
            _realise_projections(app_label, model_name)

        return view_cls

    def add_to_class(self, name, value):
        if name == "_base_manager":
            return
        super(ViewMeta, self).add_to_class(name, value)


class BaseManagerMeta:
    base_manager_name = "objects"


class View(models.Model, metaclass=ViewMeta):
    """
    Helper for exposing Postgres views as Django models.
    """

    _deferred = False
    sql = None

    @classmethod
    def get_sql(cls):
        return ViewSQL(cls.sql, None)

    @classmethod
    def get_view_connection(cls, using, restricted_mode: bool = True):
        """
        Returns connection for "using" database.
        Operates in two modes, regular mode and restricted mode.
            In regular mode just returns the connection.
            In restricted mode, returns None if migrations are not allowed (via router) to indicate view should not be
              used on the specified database.

        Overwrite this method in subclass to customize, if needed.
        """
        if not restricted_mode or router.allow_migrate(using, cls._meta.app_label):
            return connections[using]

    class Meta:
        abstract = True
        managed = False


def _realise_projections(app_label, model_name):
    """
    Checks whether the model has been loaded and runs
    realise_deferred_projections() if it has.
    """
    try:
        model_cls = apps.get_model(app_label, model_name)
    except exceptions.AppRegistryNotReady:
        return
    if model_cls is not None:
        realize_deferred_projections(model_cls)


class ReadOnlyViewQuerySet(QuerySet):
    def _raw_delete(self, *args, **kwargs):
        return 0

    def delete(self):
        raise NotImplementedError("Not allowed")

    def update(self, **kwargs):
        raise NotImplementedError("Not allowed")

    def _update(self, values):
        raise NotImplementedError("Not allowed")

    def create(self, **kwargs):
        raise NotImplementedError("Not allowed")

    def update_or_create(self, defaults=None, **kwargs):
        raise NotImplementedError("Not allowed")

    def bulk_create(self, objs, batch_size=None):
        raise NotImplementedError("Not allowed")


class ReadOnlyViewManager(models.Manager):
    def get_queryset(self):
        return ReadOnlyViewQuerySet(self.model, using=self._db)


class ReadOnlyView(View):
    """View which cannot be altered"""

    _base_manager = ReadOnlyViewManager()
    objects = ReadOnlyViewManager()

    class Meta(BaseManagerMeta):
        abstract = True
        managed = False


class MaterializedView(View):
    """A materialized view.
    More information:
    http://www.postgresql.org/docs/current/static/sql-creatematerializedview.html
    """

    with_data = True

    @classmethod
    def refresh(self, concurrently=False):
        conn = self.get_view_connection(using=router.db_for_write(self), restricted_mode=False)
        if not conn:
            logger.warning("Failed to find connection to refresh %s", self)
            return
        cursor_wrapper = conn.cursor()
        cursor = cursor_wrapper.cursor
        try:
            if self._concurrent_index is not None and concurrently:
                cursor.execute("REFRESH MATERIALIZED VIEW CONCURRENTLY {0}".format(self._meta.db_table))
            else:
                cursor.execute("REFRESH MATERIALIZED VIEW {0}".format(self._meta.db_table))
        finally:
            cursor_wrapper.close()

    class Meta:
        abstract = True
        managed = False


class ReadOnlyMaterializedView(MaterializedView):
    """Read-only version of the materialized view"""

    _base_manager = ReadOnlyViewManager()
    objects = ReadOnlyViewManager()

    class Meta(BaseManagerMeta):
        abstract = True
        managed = False



from __future__ import absolute_import, print_function, unicode_literals

from os.path import isfile

from setuptools import setup, find_packages


if isfile("README.md"):
    LONG_DESCRIPTION = open("README.md").read()
else:
    LONG_DESCRIPTION = ""


setup(
    name="django-pgviews-redux",
    version="0.9.4",
    description="Create and manage Postgres SQL Views in Django",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Mikuláš Poul",
    author_email="mikulas.poul@xelix.com",
    license="Public Domain",
    packages=find_packages(),
    url="https://github.com/xelixdev/django-pgviews-redux",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.1",
        "Framework :: Django :: 4.2",
    ],
)



import logging

from django import apps
from django.db.models import signals

logger = logging.getLogger("django_pgviews.sync_pgviews")


class ViewConfig(apps.AppConfig):
    """
    The base configuration for Django PGViews. We use this to setup our
    post_migrate signal handlers.
    """

    counter = 0
    name = "django_pgviews"
    verbose_name = "Django Postgres Views"

    def sync_pgviews(self, sender, app_config, using, **kwargs):
        """
        Forcibly sync the views.
        """
        self.counter = self.counter + 1
        total = len([a for a in apps.apps.get_app_configs() if a.models_module is not None])

        if self.counter == total:
            logger.info("All applications have migrated, time to sync")
            # Import here otherwise Django doesn't start properly
            # (models in app init are not allowed)
            from .models import ViewSyncer
            from django.conf import settings

            vs = ViewSyncer()
            vs.run(
                force=True,
                update=True,
                materialized_views_check_sql_changed=getattr(settings, "MATERIALIZED_VIEWS_CHECK_SQL_CHANGED", False),
                using=using,
            )
            self.counter = 0

    def ready(self):
        """
        Find and setup the apps to set the post_migrate hooks for.
        """
        signals.post_migrate.connect(self.sync_pgviews)



from django.db.models.sql import compiler


class NonQuotingCompiler(compiler.SQLCompiler):
    """
    Compiler for functions/statements that doesn't quote the db_table attribute.
    """

    def quote_name_unless_alias(self, name):
        """
        Don't quote the name.
        """
        if name in self.quote_cache:
            return self.quote_cache[name]

        self.quote_cache[name] = name
        return name

    def as_sql(self, *args, **kwargs):
        """
        Messy hack to create some table aliases for us.
        """
        self.query.table_map[self.query.model._meta.db_table] = [""]
        return super(NonQuotingCompiler, self).as_sql(*args, **kwargs)
